#include <stdio.h>
#include <gl.h>
#include <device.h>
#include "numerics.h"
#include "ui.h"
#include "vect.h"

char *Progname;

float Espin[4];         /* Euler parameter incremental rotation */
float Erot[4];          /* Euler parameter total rotation */
float Trans[3];         /* translation */

float Zrot = 0.0;

static Matrix Imat = {
	{ 1.0, 0.0, 0.0, 0.0 },
	{ 0.0, 1.0, 0.0, 0.0 },
	{ 0.0, 0.0, 1.0, 0.0 },
	{ 0.0, 0.0, 0.0, 1.0 },
};

Matrix Lighting_matrix;

void remember_view(float *, float *);
void draw_scene(void);
void do_windows(void);
void update_objeulers(void);
void draw_objects(void);

extern char *malloc(/* unsigned */);

main(int argc, char **argv)
{
	register char *curr_arg;
	register int c;
	int i;
	FILE *infile;

	Progname = *argv;

	/* Parse arguments beginning with a hyphen. */

	while (--argc > 0 && **++argv == '-' && *(curr_arg = 1 + *argv)) {

		/* Scan current argument until a flag indicates that the rest
		of the argument isn't flags (curr_arg = NULL), or until
		the end of the argument is reached (if it is all flags). */

		while (curr_arg != NULL && (c = *curr_arg++)) {
			switch (c) {

			default:
				fprintf(stderr, "%s: unknown flag -%c\n", Progname, c);
				exit(-1);
			}
		}
	}

	/* Set up a window. */

	foreground();
	keepaspect(1, 1);
	winopen("");
	doublebuffer();
	RGBmode();
	gconfig();
	zbuffer(TRUE);

	mmode(MVIEWING);

	/* projection matrix */

	ortho(0.0, 100.0, 0.0, 100.0, -60.0, 100.0);

	/* modelview (transformation) matrix */

	loadmatrix(Imat);
	getmatrix(Lighting_matrix);

	shademodel(GOURAUD);
	def_light_calc();
	use_light_calc(1);

	RGBcolor(255, 255, 255);
	clear();
	zclear();
	swapbuffers();

	/* Initial rotation and translation. */

	Espin[0] = Espin[1] = Espin[2] = 0.0;
	Espin[3] = 1.0;
	Erot[0] = Erot[1] = Erot[2] = 0.5;
	Erot[3] = 0.5;
	Trans[0] = Trans[1] = Trans[2] = 0.0;

	/* Queue some events, and tell ui() what to do. */

	add_event(ANY, ESCKEY, UP, ui_exit, 0);
	qdevice(ESCKEY);
	add_event(ANY, WINQUIT, ANY, ui_exit, 0);
	qdevice(WINQUIT);
	add_update(&ui_quiet, draw_scene, NULL);

	ui(remember_view);      /* Time is spent in ui, interacting */
}

/* Called by ui interface, passed a float[4] that is a rotation
specified in Euler paramaters and a float[3] that is xyz translation.
No rotation is { 0.0, 0.0, 0.0, 1.0 }, no translation is { 0.0, 0.0, 0.0 }. */

void
remember_view(float *rot, float *trans)
{
	vcopy(rot, Espin);
	Espin[3] = rot[3];
	vadd(trans, Trans, Trans);
	update_objeulers();
	do_windows();
}

/* This function is called whenever the user isn't interacting with
the program (when ui_quiet is TRUE).  It just keeps on applying the
last rotation and drawing the scene over and over. */

void
draw_scene(void)
{
	update_objeulers();
	do_windows();
}

void
do_windows()
{
	RGBcolor(255, 255, 255);
	clear();
	zclear();
	draw_objects();
	swapbuffers();
}

/* This routine updates an object's total rotation by applying its
spin rotation, and coming up with a new total rotation. */

void
update_objeulers()
{
	add_eulers(Espin, Erot, Erot);
}

/* Draw the objects. */

void
draw_objects()
{
	Matrix m;

	pushmatrix();

	/* Translate */

	translate(Trans[0] * 10.0, Trans[1] * 10.0, Trans[2] * 10.0);

	/* Rotate */

	build_rotmatrix(m, Erot);
	multmatrix(m);
	if (getbutton(ZKEY)) {
		Zrot += 5.0;
	}
	rot(Zrot, 'z');

	/* Draw the object. */

	drawit();

	popmatrix();
}

/* Material properties. */

float   MatColor[] = {
	AMBIENT,     0.400,   0.400,   0.400,
	DIFFUSE,     256.0/256.0, 250.0/256.0,  240.0/256.0,
	SPECULAR,    0.002,   0.002,   0.002,
	EMISSION,    0.000,   0.000,   0.000,
	ALPHA,       1.000,
	SHININESS,   0.000,
	LMNULL
};

float   MatGreenPlastic[] = {
	AMBIENT,     0.000,   0.000,   0.000,
	DIFFUSE,     0.100,   0.650,   0.100,
	SPECULAR,    0.350,   0.650,   0.450,
	EMISSION,    0.000,   0.000,   0.000,
	ALPHA,       1.000,
	SHININESS,   0.000,
	LMNULL
};

float   MatWhitePlastic[] = {
	AMBIENT,     0.000,   0.000,   0.000,
	DIFFUSE,     0.850,   0.850,   0.850,
	SPECULAR,    0.400,   0.400,   0.400,
	EMISSION,    0.000,   0.000,   0.000,
	ALPHA,       1.000,
	SHININESS,  20.000,
	LMNULL
};

float   MatLight[] = {
	AMBIENT,        0.0, 0.0, 0.0,
	LCOLOR,         1.0, 1.0, 1.0,
	POSITION,       0.0, 0.0, 1.0, 0.0,
	SPOTDIRECTION,  0.0, 0.0, -1.0,
	SPOTLIGHT,      0.0, 180.0,
	LMNULL
};

float   MatLight2[] = {
	AMBIENT,        0.0, 0.0, 0.0,
	LCOLOR,         1.0, 1.0, 1.0,
	POSITION,       0.0, 0.0, 1.0, 0.0,
	SPOTDIRECTION,  0.0, -1.0, 0.0,
	SPOTLIGHT,      0.0, 10.0,
	LMNULL
};

/* Define the lighting calculation. */

def_light_calc()
{
	lmdef(DEFLIGHT, 1, 0, MatLight);
	lmdef(DEFLIGHT, 2, 0, MatLight2);
	lmdef(DEFLMODEL, 1, 0, NULL);

	lmdef(DEFMATERIAL, 1, 0, MatColor);
	lmdef(DEFMATERIAL, 2, 0, MatGreenPlastic);
#ifdef notdef
	lmdef(DEFMATERIAL, 3, 0, MatGreenPlastic);
#endif
	lmdef(DEFMATERIAL, 3, 0, MatWhitePlastic);
}

/* Switch to the correct light calc. */

use_light_calc(int index)
{
	pushmatrix();
	loadmatrix(Lighting_matrix);

	lmbind(MATERIAL, index);
	lmbind(LIGHT0, 1);
	lmbind(LIGHT1, 2);
	lmbind(LMODEL, 1);

	popmatrix();
}

drawit()
{
	int i, j;
	static float v[3];

	use_light_calc(2);

	/* A plane of points. */

	bgnpoint();
	for (i = 0; i < 100; i++) {
		for (j = 0; j < 100; j++) {
			v[0] = cos(i / 100.0 * 4.0 * M_PI);
			v[1] = 0.0;
			v[2] = sin(i / 100.0 * 4.0 * M_PI);
			n3f(v);
			v[0] = i;
			v[1] = j;
			v[2] = 0.0;
			v3f(v);
		}
	}
	endpoint();

	/* A second plane of points. */

	bgnpoint();
	for (i = 0; i < 100; i++) {
		for (j = 0; j < 100; j++) {
			v[0] = cos(i / 100.0 * M_PI / 2.0);
			v[1] = 0.0;
			v[2] = sin(i / 100.0 * M_PI / 2.0);
			n3f(v);
			v[0] = i;
			v[1] = j;
			v[2] = 20.0;
			v3f(v);
		}
	}
	endpoint();
}
