/* High-res mandelbrot sets. */

#include <stdio.h>
#include <math.h>
#include <gl.h>
#include <device.h>
#include "pwin.h"
#include "event.h"

#define iterate(x, y) miterate(x, y)

int miterate(double x, double y);
int jiterate(double x, double y);

char *Progname;

void exit();
void draw_figure();

int Maxits = 256;

double Cx;
double Cy;

double Left;
double Right;
double Bottom;
double Top;
int Wide;
int Tall;
int Maporigin;

int Init;

/* Parameter window stuff. */

PWIN *Pwin;

void get_double(int argc, char **argv, double *arg);
void get_int(int argc, char **argv, int *arg);
void set_pwin_items(PWIN *);
void save_pwin();

KEYWORD Keytable[] = {
	{ "Maxits", get_int, &Maxits },
	{ "Maporigin", get_int, &Maporigin },
	{ "Cx", get_double, &Cx },
	{ "Cy", get_double, &Cy },
	{ "left", get_double, &Left },
	{ "right", get_double, &Right },
	{ "bottom", get_double, &Bottom },
	{ "top", get_double, &Top },
	{ NULL, NULL, 0 }
};
#define NPWIN (sizeof(Keytable) / sizeof(KEYWORD) - 1)

int Win;

void redraw();
void hitme();
void drag_rect();
void reset();

main(int argc, char **argv)
{
	register char *curr_arg;
	register int c;
	int i;

	Progname = *argv;

	/* Parse arguments beginning with a hyphen. */

	while (--argc > 0 && **++argv == '-' && *(curr_arg = 1 + *argv)) {

		/* Scan current argument until a flag indicates that the rest
		of the argument isn't flags (curr_arg = NULL), or until
		the end of the argument is reached (if it is all flags). */

		while (curr_arg != NULL && (c = *curr_arg++)) {
			switch (c) {

			default:
				fprintf(stderr, "%s: unknown flag -%c\n", Progname, c);
				exit(-1);
			}
		}
	}

	/* Set up a window. */

	foreground();
	Win = winopen("");
	getsize(&Wide, &Tall);

	color(WHITE);
	clear();

	/* Parameter window. */

	Left = Bottom = -2.0;
	Right = Top = 2.0;

	Maporigin = 1024;

	Cx = Cy = 0.0;

	Pwin = create_pwin(NPWIN, 40, Keytable);

	/* If he specified a parameter file, read it in. */

	if (argc && **argv != '-') {
		scan_pfile(*argv, Pwin);
	}
	set_pwin_items(Pwin);

	/* Queue some events. */

	add_event(ANY, REDRAW, Win, redraw, 0);
	qdevice(REDRAW);
	add_event(ANY, ESCKEY, UP, exit, 0);
	qdevice(ESCKEY);
	add_event(ANY, WINQUIT, ANY, exit, 0);
	qdevice(WINQUIT);
	add_event(ANY, F1KEY, UP, save_pwin, 0);
	qdevice(F1KEY);
	add_event(ANY, F2KEY, UP, reset, 0);
	qdevice(F2KEY);

	/* Watch for mouse hits. */

	add_event(Win, LEFTMOUSE, DOWN, drag_rect, 0);
	qdevice(LEFTMOUSE);
	add_event(Win, MIDDLEMOUSE, DOWN, hitme, 0);
	qdevice(MIDDLEMOUSE);

	Init = TRUE;
	while (1) {
		if (Init) {
			Init = FALSE;
			draw_figure();
		} else {
			event();
		}
	}
}

/* Draw the figure. */

void
draw_figure()
{
	double x, y;
	double x1, y1;
	double xinc, yinc;
	int row, col, i, j;
	int its;

	ortho2(0, Wide, 0, Tall);
	color(WHITE);
	clear();

	xinc = (Right - Left) / Wide;
	yinc = (Top - Bottom) / Tall;
	for (col = 0; col < Wide; col++) {
		x = Left + col * xinc;
		for (row = 0; row < Tall; row++) {
			y = Bottom + row * yinc;
			its = iterate(x, y);
			if (its == Maxits) {
				its = 0;
				break;
			}
			color(its + Maporigin);
			pnt2i(col, row);
		}
		if (qtest()) {
			event();
			if (Init) {
				break;
			}
		}
	}
}

/* Reset the items in the parameter window. */

void
set_pwin_items(PWIN *pwin)
{
	int i;
	KEYWORD *k;
	char buf[100];

	k = Keytable;
	for (i = 0; i < NPWIN; i++, k++) {
		if (k->keyfunc == (void (*)())get_double) {
			sprintf(buf, "%s %lg", k->keyword, *((double *)k->arg));
		} else if (k->keyfunc == (void (*)())get_int) {
			sprintf(buf, "%s %d", k->keyword, *((int *)k->arg));
		}
		add_pwin_item(pwin, i, buf);
	}
}

void
save_pwin()
{
	int i;
	char **pline;
	FILE *outfile;
	char buf[100];

	printf("File name? ");
	fgets(buf, 100, stdin);
	buf[strlen(buf) - 1] = '\0';
	if ((outfile = fopen(buf, "w")) == NULL) {
		fprintf(stderr, "can't write file %s\n", buf);
		return;
	}

	pline = Pwin->text;
	for (i = Pwin->n_items - 1; i >= 0; --i) {
		fputs(*pline++, outfile);
		putc('\n', outfile);
	}

	fclose(outfile);
}

/* Called from the parameter window when the user changes a parameter. */

void
get_double(int argc, char **argv, double *fp)
{
	*fp = atof(argv[1]);
	Init = TRUE;
}

void
get_int(int argc, char **argv, int *ip)
{
	*ip = atof(argv[1]);
	Init = TRUE;
}

void
reset()
{
	Left = Bottom = -2.0;
	Right = Top = 2.0;
	set_pwin_items(Pwin);
	Init = TRUE;
}

int
jiterate(double x, double y)
{
	double cx, cy, x2, y2, xy;
	int its;

	cx = Cx;
	cy = Cy;
	for (its = 0; its < Maxits; its++) {
		x2 = x * x;
		y2 = y * y;
		if (x2 + y2 > 32.0) {
			break;
		}
		xy = x * y;
		x = x2 - y2 + cx;
		y = xy + cy;
#ifdef notdef
		x = x2 - y2 + cx * cx;
		y = xy + cy * cy;
#endif
	}

	return (its);
}

int
miterate(double cx, double cy)
{
	double x, y, x2, y2, xy;
	int its;

	x = y = 0.0;
	for (its = 0; its < Maxits; its++) {
		x2 = x * x;
		y2 = y * y;
		if (x2 + y2 > 100.0) {
			break;
		}
		xy = x * y;
		x = x2 - y2 + cx;
		y = 2.0 * xy + cy;
#ifdef notdef
		x = x2 - y2 - cx + y;
		y = xy + xy - cy;
#endif
#ifdef notdef
		x = x2 - y2 - cy;
		y = xy + y2 + cx - x2;
#endif
	}

	return (its);
}

/* Middlemouse reports the coordinate, leftmouse drags a rect. */

void
hitme()
{
	int xorg, yorg;
	double xinc, yinc;
	double x, y;

	getorigin(&xorg, &yorg);

	xinc = (Right - Left) / Wide;
	yinc = (Top - Bottom) / Tall;

	x = Left + (getvaluator(MOUSEX) - xorg) * xinc;
	y = Bottom + (getvaluator(MOUSEY) - yorg) * yinc;
	printf("%f, %f\n", x, y);
}

void
redraw()
{
	reshapeviewport();
	getsize(&Wide, &Tall);
	Init = TRUE;
}

void
drag_rect()
{
	long mx, my, orx, ory, omx, omy, cx, cy, bx, by, obx, oby;
	double corner1x, corner1y, corner2x, corner2y;
	short val;

	drawmode(OVERDRAW);
	mapcolor(0, 0, 0, 0);
	mapcolor(1, 255, 255, 255);

	qreset();

	getorigin(&orx, &ory);
	omx = mx = getvaluator(MOUSEX) - orx;
	omy = my = getvaluator(MOUSEY) - ory;

	obx = bx = cx = mx;
	oby = by = cy = my;

	corner1y = ((double) my / (double) Tall) * (Top - Bottom) + Bottom;
	corner1x = ((double) mx / (double) Wide) * (Right - Left) + Left;

	color(0);
	clear();

	while (!qtest()) {
		mx = getvaluator(MOUSEX) - orx;
		my = getvaluator(MOUSEY) - ory;

		if (mx == omx && my == omy) {
			continue;
		}

#ifdef notdef
		/* preserve aspect ratio */

		if (abs(mx - cx) * Tall / Wide > abs(my - cy)) {
			bx = mx;
			by = cy + abs(mx - cx) * sgn(my - cy) * Tall / Wide;
		} else {
			by = my;
			bx = cx + abs(my - cy) * sgn(mx - cx) * Wide / Tall;
		}
#endif
		bx = mx;
		by = my;

		color(0);
		recti(cx, cy, obx, oby);

		color(1);
		recti(cx, cy, bx, by);

		obx = bx;
		oby = by;
		omx = mx;
		omy = my;

	}

	if (qtest() != LEFTMOUSE) {
		return;
	} else {
		qread(&val);
	}

	if (by != cy && bx != cx) {
		corner2y = (double)by / (double)Tall * (Top - Bottom) + Bottom;
		corner2x = (double)bx / (double)Wide * (Right - Left) + Left;

		if (corner2y < corner1y) {
			Bottom = corner2y;
			Top = corner1y;
		} else {
			Bottom = corner1y;
			Top = corner2y;
		}

		if (corner2x < corner1x) {
			Left = corner2x;
			Right = corner1x;
		} else {
			Left = corner1x;
			Right = corner2x;
		}
	}
	set_pwin_items(Pwin);
	Init = TRUE;

	color(0);
	clear();

	drawmode(NORMALDRAW);
}

sgn(i)
{
	return (i >= 0 ? (i == 0 ? 0 : 1) : -1);
}
